@@ -12,20 +12,19 @@
 #ifndef INCLUDE_WHAT_YOU_USE_IWYU_AST_UTIL_H_
 #define INCLUDE_WHAT_YOU_USE_IWYU_AST_UTIL_H_
 
+#include <functional>                   // for function
 #include <map>                          // for map
 #include <set>                          // for set
 #include <string>                       // for string
 
-#include "iwyu_port.h"  // for CHECK_
-#include "iwyu_use_flags.h"
-#include "llvm/Support/Casting.h"
-#include "clang/AST/DeclBase.h"
 #include "clang/AST/NestedNameSpecifier.h"
-#include "clang/AST/Stmt.h"
 #include "clang/AST/TemplateBase.h"
 #include "clang/AST/Type.h"
 #include "clang/AST/TypeLoc.h"
 #include "clang/Basic/SourceLocation.h"
+#include "iwyu_port.h"  // for CHECK_
+#include "iwyu_use_flags.h"
+#include "llvm/Support/Casting.h"
 
 namespace clang {
 class CXXConstructExpr;
@@ -35,19 +34,21 @@ class CXXDestructorDecl;
 class CXXMethodDecl;
 class CXXRecordDecl;
 class CallExpr;
-class CastExpr;
 class ClassTemplateDecl;
+class ClassTemplateSpecializationDecl;
+class Decl;
+class DeclContext;
 class Expr;
 class FunctionDecl;
 class NamedDecl;
 class NamespaceDecl;
+class Stmt;
 class TagDecl;
 class TemplateDecl;
 class TemplateName;
 class TranslationUnitDecl;
 class TypeDecl;
 class ValueDecl;
-struct ASTTemplateArgumentListInfo;
 }  // namespace clang
 
 namespace include_what_you_use {
@@ -547,6 +548,11 @@ clang::SourceRange GetSourceRangeOfClassDecl(const clang::Decl* decl);
 set<clang::FunctionDecl*> GetLateParsedFunctionDecls(
     clang::TranslationUnitDecl* decl);
 
+struct TemplateInstantiationData {
+  map<const clang::Type*, const clang::Type*> resugar_map;
+  set<const clang::Type*> provided_types;
+};
+
 // One can't have partial template specialization or default template
 // args for function templates, but they're complicated in their own
 // way: they can have deduced template arguments (deduced from the
@@ -579,9 +585,12 @@ set<clang::FunctionDecl*> GetLateParsedFunctionDecls(
 // NOTE: This routine is far from perfect.  To really do this right,
 // we'd need to refactor SemaTemplateDeduction to take an argument to
 // not canonicalize deduced template arguments.
+// Besides, the returned data contains a set of "provided" types (e.g. from
+// a typedef being type template argument).
 // calling_expr should be a CallExpr, CXXConstructExpr, or DeclRefExpr.
-map<const clang::Type*, const clang::Type*> GetTplTypeResugarMapForFunction(
-    const clang::FunctionDecl* decl, const clang::Expr* calling_expr);
+TemplateInstantiationData GetTplInstDataForFunction(
+    const clang::FunctionDecl* decl, const clang::Expr* calling_expr,
+    std::function<set<const clang::Type*>(const clang::Type*)> provided_getter);
 
 // If class_decl is instantiated from a class template,
 // returns the decl for that template; otherwise returns class_decl.
@@ -624,6 +633,14 @@ bool IsFriendDecl(const clang::Decl* decl);
 // or definition.
 bool IsExplicitInstantiation(const clang::Decl* decl);
 
+// If an explicit instantiation definition follows the corresponding explicit
+// instantiation declaration, clang marks the first declaration as definition.
+// This function returns true if the given specialization declaration
+// corresponds to the genuine instantiation definition, as written
+// in the source.
+bool IsExplicitInstantiationDefinitionAsWritten(
+    const clang::ClassTemplateSpecializationDecl*);
+
 // Returns true if this decl is nested inside an inline namespace.
 bool IsInInlineNamespace(const clang::Decl* decl);
 
@@ -812,17 +829,21 @@ bool HasImplicitConversionConstructor(const clang::Type* type);
 // We also include mappings for component types: if we have an entry
 // 'vector<TypedefType>' -> 'vector<Foo>', we also add an entry
 // 'TypedefType' -> 'Foo'.
+// Besides, the returned data contains a set of "provided" types (e.g. from
+// a typedef being type template argument).
 // For ease of calling, this accept any type, but will return an empty
-// map for any input that's not a template specialization type.
-map<const clang::Type*, const clang::Type*> GetTplTypeResugarMapForClass(
-    const clang::Type* type);
+// result for any input that's not a template specialization type.
+TemplateInstantiationData GetTplInstDataForClass(
+    const clang::Type* type,
+    std::function<set<const clang::Type*>(const clang::Type*)> provided_getter);
 
-// Like GetTplTypeResugarMapForClass, but if a type has
+// Like GetTplInstDataForClass, but if a type has
 // components (for instance, 'Foo*' and 'vector<Foo>' both
 // have a component Foo), we don't include the components
 // in the result-map.
-map<const clang::Type*, const clang::Type*>
-GetTplTypeResugarMapForClassNoComponentTypes(const clang::Type* type);
+TemplateInstantiationData GetTplInstDataForClassNoComponentTypes(
+    const clang::Type* type,
+    std::function<set<const clang::Type*>(const clang::Type*)> provided_getter);
 
 // Returns true if, for the given enumeration type, opaque (i.e. forward,
 // in fact) declarations are allowed. It means that the enumeration should be
@@ -884,6 +905,9 @@ std::string GetKindName(const clang::TypeLoc typeloc);
 // visible from said function.
 bool IsDeclaredInsideFunction(const clang::Decl* decl);
 
+// Returns true if decl is partially inside a macro.
+bool IsDeclaredInsideMacro(const clang::Decl* decl);
+
 }  // namespace include_what_you_use
 
 #endif  // INCLUDE_WHAT_YOU_USE_IWYU_AST_UTIL_H_
