@@ -11,16 +11,16 @@
 
 #include <algorithm>                    // for find
 #include <cstddef>                      // for size_t
-// TODO(wan): make sure IWYU doesn't suggest <iterator>.
-#include <iterator>                     // for find
 // not hash_map: it's not as portable and needs hash<string>.
 #include <map>                          // for map, map<>::mapped_type, etc
 #include <memory>
+#include <numeric>                      // for accumulate
 #include <string>                       // for string, basic_string, etc
 #include <system_error>                 // for error_code
 #include <utility>                      // for pair, make_pair
 #include <vector>                       // for vector, vector<>::iterator
 
+#include "clang/Tooling/Inclusions/StandardLibrary.h"
 #include "iwyu_location_util.h"
 #include "iwyu_path_util.h"
 #include "iwyu_port.h"
@@ -28,25 +28,23 @@
 #include "iwyu_stl_util.h"
 #include "iwyu_string_util.h"
 #include "iwyu_verrs.h"
-
+#include "llvm/ADT/STLExtras.h"
+#include "llvm/ADT/SmallString.h"
+#include "llvm/ADT/SmallVector.h"
 #include "llvm/ADT/StringRef.h"
+#include "llvm/ADT/Twine.h"
 #include "llvm/Support/Casting.h"
 #include "llvm/Support/ErrorOr.h"
 #include "llvm/Support/FileSystem.h"
+#include "llvm/Support/Format.h"
 #include "llvm/Support/MemoryBuffer.h"
 #include "llvm/Support/SourceMgr.h"
 #include "llvm/Support/YAMLParser.h"
-#include "clang/Basic/FileManager.h"
-#include "clang/Tooling/Inclusions/StandardLibrary.h"
 
-using std::find;
-using std::make_pair;
-using std::map;
-using std::pair;
-using std::string;
-using std::unique_ptr;
-using std::vector;
+// TODO: Clean out pragmas as IWYU improves.
+// IWYU pragma: no_include <iterator>
 
+using clang::OptionalFileEntryRef;
 using llvm::MemoryBuffer;
 using llvm::SourceMgr;
 using llvm::yaml::KeyValueNode;
@@ -56,6 +54,13 @@ using llvm::yaml::ScalarNode;
 using llvm::yaml::SequenceNode;
 using llvm::yaml::Stream;
 using llvm::yaml::document_iterator;
+using std::find;
+using std::make_pair;
+using std::map;
+using std::pair;
+using std::string;
+using std::unique_ptr;
+using std::vector;
 
 namespace include_what_you_use {
 
@@ -114,7 +119,6 @@ const IncludeMapEntry libc_symbol_map[] = {
   { "fsfilcnt_t", kPrivate, "<sys/types.h>", kPublic },
   { "getopt", kPrivate, "<unistd.h>", kPublic },
   { "gid_t", kPrivate, "<sys/types.h>", kPublic },
-  { "gid_t", kPrivate, "<unistd.h>", kPublic },
   { "htonl", kPrivate, "<arpa/inet.h>", kPublic },
   { "htons", kPrivate, "<arpa/inet.h>", kPublic },
   { "id_t", kPrivate, "<sys/types.h>", kPublic },
@@ -122,7 +126,6 @@ const IncludeMapEntry libc_symbol_map[] = {
   { "intmax_t", kPrivate, "<stdint.h>", kPublic },
   { "uintmax_t", kPrivate, "<stdint.h>", kPublic },
   { "ino64_t", kPrivate, "<sys/types.h>", kPublic },
-  { "ino64_t", kPrivate, "<dirent.h>", kPublic },
   { "ino_t", kPrivate, "<sys/types.h>", kPublic },
   { "int8_t", kPrivate, "<stdint.h>", kPublic },
   { "int16_t", kPrivate, "<stdint.h>", kPublic },
@@ -146,15 +149,12 @@ const IncludeMapEntry libc_symbol_map[] = {
   { "ntohl", kPrivate, "<arpa/inet.h>", kPublic },
   { "ntohs", kPrivate, "<arpa/inet.h>", kPublic },
   { "off64_t", kPrivate, "<sys/types.h>", kPublic },
-  { "off64_t", kPrivate, "<unistd.h>", kPublic },
   { "off_t", kPrivate, "<sys/types.h>", kPublic },
-  { "off_t", kPrivate, "<unistd.h>", kPublic },
   { "optarg", kPrivate, "<unistd.h>", kPublic },
   { "opterr", kPrivate, "<unistd.h>", kPublic },
   { "optind", kPrivate, "<unistd.h>", kPublic },
   { "optopt", kPrivate, "<unistd.h>", kPublic },
   { "pid_t", kPrivate, "<sys/types.h>", kPublic },
-  { "pid_t", kPrivate, "<unistd.h>", kPublic },
   { "ptrdiff_t", kPrivate, "<stddef.h>", kPublic },
   { "regex_t", kPrivate, "<regex.h>", kPublic },
   { "regmatch_t", kPrivate, "<regex.h>", kPublic },
@@ -166,7 +166,6 @@ const IncludeMapEntry libc_symbol_map[] = {
   { "sockaddr", kPrivate, "<sys/socket.h>", kPublic },
   { "socklen_t", kPrivate, "<sys/socket.h>", kPublic },
   { "ssize_t", kPrivate, "<sys/types.h>", kPublic },
-  { "ssize_t", kPrivate, "<unistd.h>", kPublic },
   { "stat", kPrivate, "<sys/stat.h>", kPublic },
   { "suseconds_t", kPrivate, "<sys/types.h>", kPublic },
   { "time_t", kPrivate, "<time.h>", kPublic },
@@ -178,9 +177,7 @@ const IncludeMapEntry libc_symbol_map[] = {
   { "tm", kPrivate, "<time.h>", kPublic },
   { "u_char", kPrivate, "<sys/types.h>", kPublic },
   { "uid_t", kPrivate, "<sys/types.h>", kPublic },
-  { "uid_t", kPrivate, "<unistd.h>", kPublic },
   { "useconds_t", kPrivate, "<sys/types.h>", kPublic },
-  { "useconds_t", kPrivate, "<unistd.h>", kPublic },
   { "wchar_t", kPrivate, "<stddef.h>", kPublic },
   { "wchar_t", kPrivate, "<stdlib.h>", kPublic },
   // It is unspecified if the cname headers provide ::size_t.
@@ -190,10 +187,8 @@ const IncludeMapEntry libc_symbol_map[] = {
   { "size_t", kPrivate, "<stdio.h>", kPublic },
   { "size_t", kPrivate, "<stdlib.h>", kPublic },
   { "size_t", kPrivate, "<string.h>", kPublic },
-  { "size_t", kPrivate, "<sys/types.h>", kPublic },
   { "size_t", kPrivate, "<time.h>", kPublic },
   { "size_t", kPrivate, "<uchar.h>", kPublic },
-  { "size_t", kPrivate, "<unistd.h>", kPublic },
   { "size_t", kPrivate, "<wchar.h>", kPublic },
   // Macros that can be defined in more than one file, don't have the
   // same __foo_defined guard that other types do, so the grep above
@@ -233,7 +228,6 @@ const IncludeMapEntry libc_symbol_map[] = {
   { "NULL", kPrivate, "<stdlib.h>", kPublic },
   { "NULL", kPrivate, "<string.h>", kPublic },
   { "NULL", kPrivate, "<time.h>", kPublic },
-  { "NULL", kPrivate, "<unistd.h>", kPublic },
   { "NULL", kPrivate, "<wchar.h>", kPublic },
   { "offsetof", kPrivate, "<stddef.h>", kPublic },
 };
@@ -783,6 +777,52 @@ const IncludeMapEntry libstdcpp_include_map[] = {
   { "<bits/valarray_array.tcc>", kPrivate, "<valarray>", kPublic },
   { "<bits/valarray_before.h>", kPrivate, "<valarray>", kPublic },
   { "<bits/vector.tcc>", kPrivate, "<vector>", kPublic },
+  // I don't have a reusable pipeline for the debug headers, but they're pretty
+  // straightforward (<debug/xx> maps to <xx>, and then private-to-private for
+  // the rest).
+  { "<debug/bitset>", kPrivate, "<bitset>", kPublic },
+  { "<debug/deque>", kPrivate, "<deque>", kPublic },
+  { "<debug/forward_list>", kPrivate, "<forward_list>", kPublic },
+  { "<debug/list>", kPrivate, "<list>", kPublic },
+  { "<debug/map.h>", kPrivate, "<debug/map>", kPrivate },
+  { "<debug/map>", kPrivate, "<map>", kPublic },
+  { "<debug/multimap.h>", kPrivate, "<debug/map>", kPrivate },
+  { "<debug/multiset.h>", kPrivate, "<debug/set>", kPrivate },
+  { "<debug/safe_container.h>", kPrivate, "<debug/deque>", kPrivate },
+  { "<debug/safe_container.h>", kPrivate, "<debug/forward_list>", kPrivate },
+  { "<debug/safe_container.h>", kPrivate, "<debug/list>", kPrivate },
+  { "<debug/safe_container.h>", kPrivate, "<debug/string>", kPrivate },
+  { "<debug/safe_container.h>", kPrivate, "<debug/unordered_map>", kPrivate },
+  { "<debug/safe_container.h>", kPrivate, "<debug/unordered_set>", kPrivate },
+  { "<debug/safe_container.h>", kPrivate, "<debug/vector>", kPrivate },
+  { "<debug/safe_iterator.h>", kPrivate, "<debug/bitset>", kPrivate },
+  { "<debug/safe_iterator.h>", kPrivate, "<debug/deque>", kPrivate },
+  { "<debug/safe_iterator.h>", kPrivate, "<debug/forward_list>", kPrivate },
+  { "<debug/safe_iterator.h>", kPrivate, "<debug/list>", kPrivate },
+  { "<debug/safe_iterator.h>", kPrivate, "<debug/string>", kPrivate },
+  { "<debug/safe_iterator.h>", kPrivate, "<debug/unordered_map>", kPrivate },
+  { "<debug/safe_iterator.h>", kPrivate, "<debug/unordered_set>", kPrivate },
+  { "<debug/safe_iterator.h>", kPrivate, "<debug/vector>", kPrivate },
+  { "<debug/safe_iterator.tcc>", kPrivate, "<debug/safe_iterator.h>", kPrivate },
+  { "<debug/safe_local_iterator.h>", kPrivate, "<debug/unordered_map>", kPrivate },
+  { "<debug/safe_local_iterator.h>", kPrivate, "<debug/unordered_set>", kPrivate },
+  { "<debug/safe_local_iterator.tcc>", kPrivate, "<debug/safe_local_iterator.h>", kPrivate },
+  { "<debug/safe_sequence.h>", kPrivate, "<debug/bitset>", kPrivate },
+  { "<debug/safe_sequence.h>", kPrivate, "<debug/deque>", kPrivate },
+  { "<debug/safe_sequence.h>", kPrivate, "<debug/forward_list>", kPrivate },
+  { "<debug/safe_sequence.h>", kPrivate, "<debug/list>", kPrivate },
+  { "<debug/safe_sequence.h>", kPrivate, "<debug/string>", kPrivate },
+  { "<debug/safe_sequence.h>", kPrivate, "<debug/vector>", kPrivate },
+  { "<debug/safe_sequence.tcc>", kPrivate, "<debug/safe_sequence.h>", kPrivate },
+  { "<debug/safe_unordered_container.h>", kPrivate, "<debug/unordered_map>", kPrivate },
+  { "<debug/safe_unordered_container.h>", kPrivate, "<debug/unordered_set>", kPrivate },
+  { "<debug/safe_unordered_container.tcc>", kPrivate, "<debug/safe_unordered_container.h>", kPrivate },
+  { "<debug/set.h>", kPrivate, "<debug/set>", kPrivate },
+  { "<debug/set>", kPrivate, "<set>", kPublic },
+  { "<debug/string>", kPrivate, "<string>", kPublic },
+  { "<debug/unordered_map>", kPrivate, "<unordered_map>", kPublic },
+  { "<debug/unordered_set>", kPrivate, "<unordered_set>", kPublic },
+  { "<debug/vector>", kPrivate, "<vector>", kPublic },
   { "<decimal/decimal.h>", kPrivate, "<decimal>", kPublic },
   { "<experimental/bits/fs_dir.h>", kPrivate, "<experimental/filesystem>", kPublic },
   { "<experimental/bits/fs_fwd.h>", kPrivate, "<experimental/filesystem>", kPublic },
@@ -911,7 +951,6 @@ const IncludeMapEntry libstdcpp_include_map[] = {
   // ( cd /usr/crosstool/v12/gcc-4.3.1-glibc-2.3.6-grte/x86_64-unknown-linux-gnu/x86_64-unknown-linux-gnu/include/c++/4.3.1 && grep -R '^ *# *include.*tcc' * | perl -nle 'm/^([^:]+).*[<"]([^>"]+)[>"]/ && print qq@    { "<$2>", kPrivate, "<$1>", kPublic },@' | sort )
   // I had to manually edit some of the entries to say the map-to is private.
   { "<bits/cmath.tcc>", kPrivate, "<cmath>", kPublic },
-  { "<debug/safe_iterator.tcc>", kPrivate, "<debug/safe_iterator.h>", kPublic },
   { "<tr1_impl/random.tcc>", kPrivate, "<tr1_impl/random>", kPrivate },
   // Some bits->bits #includes: A few files in bits re-export
   // symbols from other files in bits.
@@ -970,17 +1009,18 @@ const IncludeMapEntry libcxx_include_map[] = {
     // For the following entries:
     // cd llvm-project/libcxx/include ; find -type d -name "__*" | sort | sed -e "s#./__\(.*\)#  { \"@<__\1/.*>\", kPrivate, \"<\1>\", kPublic },#"
     //
-    // tweak tuple_dir entry, and comment out debug_utils, fwd, support
+    // tweak locale_dir entry, and comment out debug_utils, fwd, pstl, support
     {"@<__algorithm/.*>", kPrivate, "<algorithm>", kPublic},
     {"@<__atomic/.*>", kPrivate, "<atomic>", kPublic},
-    {"@<__algorithm/.*>", kPrivate, "<algorithm>", kPublic},
     {"@<__bit/.*>", kPrivate, "<bit>", kPublic},
     {"@<__charconv/.*>", kPrivate, "<charconv>", kPublic},
     {"@<__chrono/.*>", kPrivate, "<chrono>", kPublic},
     {"@<__compare/.*>", kPrivate, "<compare>", kPublic},
     {"@<__concepts/.*>", kPrivate, "<concepts>", kPublic},
+    {"@<__condition_variable/.*>", kPrivate, "<condition_variable>", kPublic},
     {"@<__coroutine/.*>", kPrivate, "<coroutine>", kPublic},
     //{ "@<__debug_utils/.*>", kPrivate, "<debug_utils>", kPublic },
+    {"@<__exception/.*>", kPrivate, "<exception>", kPublic},
     {"@<__expected/.*>", kPrivate, "<expected>", kPublic},
     {"@<__filesystem/.*>", kPrivate, "<filesystem>", kPublic},
     {"@<__format/.*>", kPrivate, "<format>", kPublic},
@@ -988,21 +1028,30 @@ const IncludeMapEntry libcxx_include_map[] = {
     //{ "@<__fwd/.*>", kPrivate, "<fwd>", kPublic },
     {"@<__ios/.*>", kPrivate, "<ios>", kPublic},
     {"@<__iterator/.*>", kPrivate, "<iterator>", kPublic},
+    {"@<__locale_dir/.*>", kPrivate, "<locale>", kPublic},
+    {"@<__mdspan/.*>", kPrivate, "<mdspan>", kPublic},
     {"@<__memory/.*>", kPrivate, "<memory>", kPublic},
     {"@<__memory_resource/.*>", kPrivate, "<memory_resource>", kPublic},
+    {"@<__mutex/.*>", kPrivate, "<mutex>", kPublic},
     {"@<__numeric/.*>", kPrivate, "<numeric>", kPublic},
+    //{"@<__pstl/.*>", kPrivate, "<pstl>", kPublic},
     {"@<__random/.*>", kPrivate, "<random>", kPublic},
     {"@<__ranges/.*>", kPrivate, "<ranges>", kPublic},
+    {"@<__stop_token/.*>", kPrivate, "<stop_token>", kPublic},
     {"@<__string/.*>", kPrivate, "<string>", kPublic},
     //{ "@<__support/.*>", kPrivate, "<support>", kPublic },
+    {"@<__system_error/.*>", kPrivate, "<system_error>", kPublic},
     {"@<__thread/.*>", kPrivate, "<thread>", kPublic},
-    {"@<__tuple_dir/.*>", kPrivate, "<tuple>", kPublic},
+    {"@<__tuple/.*>", kPrivate, "<tuple>", kPublic},
     {"@<__type_traits/.*>", kPrivate, "<type_traits>", kPublic},
     {"@<__utility/.*>", kPrivate, "<utility>", kPublic},
     {"@<__variant/.*>", kPrivate, "<variant>", kPublic},
 
     // For older MacOS libc++ (13.0.0), on macOS Ventura (13.2.1)
     {"<__functional_base>", kPrivate, "<functional>", kPublic},
+
+    // For older libc++ (16.x)
+    {"@<__tuple_dir/.*>", kPrivate, "<tuple>", kPublic},
 };
 
 // Returns true if str is a valid quoted filepath pattern (i.e. either
@@ -1175,6 +1224,43 @@ string FindFileInSearchPath(const vector<string>& search_path,
   return filename;
 }
 
+void PrintMappings(const IncludePicker::IncludeMap& map, const char* name) {
+  if (map.empty()) {
+    llvm::errs() << name << ": empty, no statistics\n";
+    return;
+  }
+
+  // Compute and print count/min/max/average.
+  std::vector<size_t> lengths;
+  for (const auto& m : map) {
+    lengths.push_back(m.second.size());
+  }
+
+  size_t count = lengths.size();
+  size_t min = *std::min_element(lengths.begin(), lengths.end());
+  size_t max = *std::max_element(lengths.begin(), lengths.end());
+  double avg = std::accumulate(lengths.begin(), lengths.end(), 0.0) / count;
+  llvm::errs() << name << ": count=" << count << ", min=" << min
+               << ", max=" << max << ", average=" << llvm::format("%.2f", avg)
+               << "\n";
+
+  // Sort all mappings by size, ascending, and print them.
+  vector<pair<string, vector<MappedInclude>>> mappings(map.begin(), map.end());
+  std::stable_sort(mappings.begin(), mappings.end(), [](auto& lhs, auto& rhs) {
+    return rhs.second.size() < lhs.second.size();
+  });
+
+  llvm::errs() << "---\n";
+  for (const auto& m : mappings) {
+    llvm::errs() << m.first << ": [";
+    llvm::interleaveComma(m.second, llvm::errs(), [](const MappedInclude& x) {
+      llvm::errs() << x.quoted_include;
+    });
+    llvm::errs() << "]\n";
+  }
+  llvm::errs() << "---\n";
+}
+
 }  // anonymous namespace
 
 MappedInclude::MappedInclude(const string& q, const string& p)
@@ -1488,6 +1574,12 @@ void IncludePicker::FinalizeAddedIncludes() {
   }
 
   has_called_finalize_added_include_lines_ = true;
+
+  // Print some mapping statistics.
+  if (ShouldPrint(9)) {
+    PrintMappings(filepath_include_map_, "filepath_include_map_");
+    PrintMappings(symbol_include_map_, "symbol_include_map_");
+  }
 }
 
 // For the given key, return the vector of values associated with that
@@ -1641,7 +1733,7 @@ bool IncludePicker::HasMapping(const string& map_from_filepath,
   return quoted_to == quoted_from;   // indentity mapping, why not?
 }
 
-bool IncludePicker::IsPublic(const clang::FileEntry* file) const {
+bool IncludePicker::IsPublic(OptionalFileEntryRef file) const {
   CHECK_(file && "Need existing FileEntry");
   const string path = GetFilePath(file);
   const string quoted_file = ConvertToQuotedInclude(path);
