@@ -44,15 +44,17 @@
 #ifndef INCLUDE_WHAT_YOU_USE_IWYU_INCLUDE_PICKER_H_
 #define INCLUDE_WHAT_YOU_USE_IWYU_INCLUDE_PICKER_H_
 
+#include <cstddef>
 #include <map>                          // for map, map<>::value_compare
 #include <set>                          // for set
 #include <string>                       // for string
 #include <utility>                      // for pair
 #include <vector>                       // for vector
 
-namespace clang {
-class FileEntry;
-}  // namespace clang
+#include "clang/Basic/FileEntry.h"
+
+// TODO: Clean out pragmas as IWYU improves.
+// IWYU pragma: no_include <iterator>
 
 namespace include_what_you_use {
 
@@ -63,9 +65,9 @@ using std::string;
 
 using std::vector;
 
+enum class RegexDialect;
 struct IncludeMapEntry;
 
-enum class RegexDialect;
 enum IncludeVisibility { kUnusedVisibility, kPublic, kPrivate };
 enum class CStdLib { None, ClangSymbols, Glibc };
 enum class CXXStdLib { None, ClangSymbols, Libstdcxx, Libcxx };
@@ -180,7 +182,7 @@ class IncludePicker {
   bool HasMapping(const string& map_from_filepath,
                   const string& map_to_filepath) const;
 
-  bool IsPublic(const clang::FileEntry* file) const;
+  bool IsPublic(clang::OptionalFileEntryRef file) const;
 
   // Parses a YAML/JSON file containing mapping directives of various types.
   void AddMappingsFromFile(const string& filename);
