@@ -9,8 +9,6 @@
 
 #include "iwyu_location_util.h"
 
-#include "iwyu_ast_util.h"
-#include "iwyu_port.h"
 #include "clang/AST/Decl.h"
 #include "clang/AST/DeclBase.h"
 #include "clang/AST/DeclCXX.h"
@@ -22,6 +20,11 @@
 #include "clang/AST/TemplateBase.h"
 #include "clang/AST/TypeLoc.h"
 #include "clang/Basic/SourceLocation.h"
+#include "iwyu_ast_util.h"
+#include "iwyu_string_util.h"
+
+// TODO: Clean out pragmas as IWYU improves.
+// IWYU pragma: no_include "clang/Basic/CustomizableOptional.h"
 
 using clang::BinaryOperator;
 using clang::CXXDependentScopeMemberExpr;
@@ -29,10 +32,17 @@ using clang::CXXMethodDecl;
 using clang::CXXOperatorCallExpr;
 using clang::ClassTemplateSpecializationDecl;
 using clang::ConditionalOperator;
-using clang::FileEntry;
+using clang::Decl;
+using clang::FileID;
 using clang::FunctionDecl;
 using clang::MemberExpr;
+using clang::NestedNameSpecifierLoc;
+using clang::OptionalFileEntryRef;
 using clang::SourceLocation;
+using clang::SourceManager;
+using clang::Stmt;
+using clang::TemplateArgumentLoc;
+using clang::TypeLoc;
 using clang::UnaryOperator;
 using clang::UnresolvedMemberExpr;
 
@@ -63,7 +73,7 @@ namespace include_what_you_use {
 // methods don't have their own location anyway.
 //    Note the two issues can both be present, if an implicit method's
 // parent is an implicit instantiation.
-SourceLocation GetLocation(const clang::Decl* decl) {
+SourceLocation GetLocation(const Decl* decl) {
   if (decl == nullptr)
     return SourceLocation();
 
@@ -120,7 +130,7 @@ static SourceLocation GetMemberExprLocation(const MemberExpr* member_expr) {
   return GetInstantiationLoc(member_start);
 }
 
-SourceLocation GetLocation(const clang::Stmt* stmt) {
+SourceLocation GetLocation(const Stmt* stmt) {
   if (stmt == nullptr)
     return SourceLocation();
   // For some expressions, we take the location to be the 'key' part
@@ -153,19 +163,19 @@ SourceLocation GetLocation(const clang::Stmt* stmt) {
   return stmt->getBeginLoc();
 }
 
-SourceLocation GetLocation(const clang::TypeLoc* typeloc) {
+SourceLocation GetLocation(const TypeLoc* typeloc) {
   if (typeloc == nullptr)
     return SourceLocation();
   return typeloc->getBeginLoc();
 }
 
-SourceLocation GetLocation(const clang::NestedNameSpecifierLoc* nnsloc) {
+SourceLocation GetLocation(const NestedNameSpecifierLoc* nnsloc) {
   if (nnsloc == nullptr)
     return SourceLocation();
   return nnsloc->getBeginLoc();
 }
 
-SourceLocation GetLocation(const clang::TemplateArgumentLoc* argloc) {
+SourceLocation GetLocation(const TemplateArgumentLoc* argloc) {
   if (argloc == nullptr)
     return SourceLocation();
   return argloc->getLocation();
@@ -175,14 +185,21 @@ bool IsInScratchSpace(SourceLocation loc) {
   return StartsWith(PrintableLoc(GetSpellingLoc(loc)), "<scratch space>");
 }
 
-bool IsInHeader(const clang::Decl* decl) {
-  const FileEntry* containing_file = GetFileEntry(decl);
+bool IsInHeader(const Decl* decl) {
+  OptionalFileEntryRef containing_file = GetFileEntry(decl);
   if (!containing_file) {
     // This is a builtin, or something is terribly wrong.
     // At any rate, we're not in a header.
     return false;
   }
-  return !GlobalSourceManager()->isMainFile(*containing_file);
+  return !GlobalSourceManager()->isMainFile(containing_file->getFileEntry());
+}
+
+bool IsSystemHeader(OptionalFileEntryRef file) {
+  const SourceManager* sm = GlobalSourceManager();
+  FileID file_id = sm->translateFile(*file);
+  SourceLocation loc = sm->getLocForStartOfFile(file_id);
+  return sm->isInSystemHeader(loc);
 }
 
 }  // namespace include_what_you_use
