@@ -35,6 +35,7 @@ import sys
 import json
 import time
 import shlex
+import shutil
 import argparse
 import tempfile
 import subprocess
@@ -212,21 +213,20 @@ def split_command(cmdstr):
 
 def find_include_what_you_use():
     """ Find IWYU executable and return its full pathname. """
-    if 'IWYU_BINARY' in os.environ:
-        return os.environ.get('IWYU_BINARY')
+    env_iwyu_path = os.environ.get('IWYU_BINARY')
+    if env_iwyu_path:
+        return os.path.realpath(env_iwyu_path)
 
-    # TODO: Investigate using shutil.which when Python 2 has passed away.
-    executable_name = 'include-what-you-use'
-    if sys.platform.startswith('win'):
-        executable_name += '.exe'
+    # Search in same dir as this script.
+    iwyu_path = shutil.which('include-what-you-use',
+                             path=os.path.dirname(__file__))
+    if iwyu_path:
+        return os.path.realpath(iwyu_path)
 
-    search_path = [os.path.dirname(__file__)]
-    search_path += os.environ.get('PATH', '').split(os.pathsep)
-
-    for dirpath in search_path:
-        full = os.path.join(dirpath, executable_name)
-        if os.path.isfile(full):
-            return os.path.realpath(full)
+    # Search the system PATH.
+    iwyu_path = shutil.which('include-what-you-use')
+    if iwyu_path:
+        return os.path.realpath(iwyu_path)
 
     return None
 
@@ -360,6 +360,26 @@ def slice_compilation_db(compilation_db, selection):
     return new_db
 
 
+def worst_exit_code(worst, cur):
+    """Return the most extreme exit code of two.
+
+    Negative exit codes occur if the program exits due to a signal (Unix) or
+    structured exception (Windows). If we've seen a negative one before, keep
+    it, as it usually indicates a critical error.
+
+    Otherwise return the biggest positive exit code.
+    """
+    if cur < 0:
+        # Negative results take precedence, return the minimum
+        return min(worst, cur)
+    elif worst < 0:
+        # We know cur is non-negative, negative worst must be minimum
+        return worst
+    else:
+        # We know neither are negative, return the maximum
+        return max(worst, cur)
+
+
 def execute(invocations, verbose, formatter, jobs, max_load_average=0):
     """ Launch processes described by invocations. """
     exit_code = 0
@@ -367,7 +387,7 @@ def execute(invocations, verbose, formatter, jobs, max_load_average=0):
         for invocation in invocations:
             proc = invocation.start(verbose)
             print(formatter(proc.get_output()))
-            exit_code = max(exit_code, proc.returncode)
+            exit_code = worst_exit_code(exit_code, proc.returncode)
         return exit_code
 
     pending = []
@@ -377,7 +397,7 @@ def execute(invocations, verbose, formatter, jobs, max_load_average=0):
         for proc in complete:
             pending.remove(proc)
             print(formatter(proc.get_output()))
-            exit_code = max(exit_code, proc.returncode)
+            exit_code = worst_exit_code(exit_code, proc.returncode)
 
         # Schedule new processes if there's room.
         capacity = jobs - len(pending)
