@@ -7,9 +7,11 @@
 //
 //===----------------------------------------------------------------------===//
 
-#include "explicit_instantiation-template_direct.h"
-#include "explicit_instantiation2-template_helpers.h"
-#include "explicit_instantiation2-template_direct.h"
+// IWYU_ARGS: -I .
+
+#include "tests/cxx/explicit_instantiation-template_direct.h"
+#include "tests/cxx/explicit_instantiation2-template_helpers.h"
+#include "tests/cxx/explicit_instantiation2-template_direct.h"
 
 // These tests verify that a dependent explicit instantiation declaration is
 // always required when the main template is also needed, but not otherwise.
@@ -88,18 +90,18 @@ TemplateAsDefaultFullProvided<> t11; // 11
 /**** IWYU_SUMMARY
 
 tests/cxx/explicit_instantiation2.cc should add these lines:
-#include "explicit_instantiation-template.h"
-#include "explicit_instantiation2-template_bool.h"
-#include "explicit_instantiation2-template_short.h"
+#include "tests/cxx/explicit_instantiation-template.h"
+#include "tests/cxx/explicit_instantiation2-template_bool.h"
+#include "tests/cxx/explicit_instantiation2-template_short.h"
 
 tests/cxx/explicit_instantiation2.cc should remove these lines:
-- #include "explicit_instantiation-template_direct.h"  // lines XX-XX
-- #include "explicit_instantiation2-template_direct.h"  // lines XX-XX
+- #include "tests/cxx/explicit_instantiation-template_direct.h"  // lines XX-XX
+- #include "tests/cxx/explicit_instantiation2-template_direct.h"  // lines XX-XX
 
 The full include-list for tests/cxx/explicit_instantiation2.cc:
-#include "explicit_instantiation-template.h"  // for Template
-#include "explicit_instantiation2-template_bool.h"  // for Template
-#include "explicit_instantiation2-template_helpers.h"  // for FullUseArg, FwdDeclUseArg, TemplateAsDefaultFull, TemplateAsDefaultFullProvided, TemplateAsDefaultFwd, TemplateTemplateArgShortFull, TemplateTemplateArgShortFwd
-#include "explicit_instantiation2-template_short.h"  // for Template
+#include "tests/cxx/explicit_instantiation-template.h"  // for Template
+#include "tests/cxx/explicit_instantiation2-template_bool.h"  // for Template
+#include "tests/cxx/explicit_instantiation2-template_helpers.h"  // for FullUseArg, FwdDeclUseArg, TemplateAsDefaultFull, TemplateAsDefaultFullProvided, TemplateAsDefaultFwd, TemplateTemplateArgShortFull, TemplateTemplateArgShortFwd
+#include "tests/cxx/explicit_instantiation2-template_short.h"  // for Template
 
 ***** IWYU_SUMMARY */
